import asyncio
import re

from gmqtt import Client as MQTTClient
from gmqtt.mqtt.constants import MQTTv311

from logging import getLogger
_log = getLogger(__name__)

_instance = None


def get_mqtt():
    global _instance
    return _instance


def setup_mqtt(app=None, options=None):
    global _instance
    if _instance is not None:
        del _instance
    _instance = Mqtt(app=app, options=options)
    return _instance


class Mqtt:
    def __init__(self, app=None, options=None):
        global _instance
        if _instance is None:
            _instance = self

        self._options = options
        self.client = MQTTClient(self._options.mqtt_client_id)
        self.client.on_connect = self._on_connect
        self.client.on_message = self._on_message
        self.client.on_disconnect = self._on_disconnect
        self.client.on_subscribe = self._on_subscribe
        self._subscriptions = {}

        app.cleanup_ctx.append(self._mqtt_engine)

    async def _mqtt_engine(self, app):
        task = asyncio.create_task(self._startup(app))
        yield
        # TODO _shutdown should not be an awaitable
        await self._shutdown(app, task)

    async def _startup(self, app) -> None:
        _log.debug('MQTT Engine startup called')
        await self.client.connect(self._options.mqtt_broker_host,
                                  port=self._options.mqtt_broker_port,
                                  keepalive=60,
                                  version=MQTTv311)
        return None

    async def _shutdown(self, app, task) -> None:
        _log.debug('MQTT Engine shutdown called')
        # TODO hier kein awaitable sinnvoll
        await self.client.disconnect()
        return None

    def _on_connect(self, client, flags, rc, properties):
        _log.debug('MQTT Connected')

        for key in self._subscriptions:
            self.client.subscribe(key, qos=0)

    async def _on_message(self, client, topic, payload, qos, properties):
        payload = payload.decode()
#        _log.debug(f'MQTT RECV MSG: {payload}, TOPIC: {topic}')
        for key in self._subscriptions:
            if self.match_topic(topic, key):
                func = self._subscriptions[key]
                try:
                    await func(topic, payload)
                except Exception as e:
                    _log.error(
                        f'{__name__}, exeption {e} in callable {func}')
        return 0

    def _on_disconnect(self, client, packet, exc=None):
        _log.debug('MQTT Disconnected')

    def _on_subscribe(self, client, mid, qos, properties):
        _log.debug('MQTT SUBSCRIBED')

    def subscribe(self, topic: str, callback: callable) -> None:
        self._subscriptions[topic] = callback
# TODO only subscription from callbacks are possible,
# because client is not connected in this momen

#    Der MQTT Client ist noch nicht gestartet.
#    Subscribe ist noch nicht möglich.
#        self.client.subscribe(topic, qos=0)
        return None

    def match_topic(self, topic: str, sub: str) -> bool:
        # Achtung auch "stat/+/RESULT" und "stat/+/RESULT1234"
        # ergebn ein Match! Die regex ist nicht korrekt!
        ret = bool(re.match(sub.translate({43: "[^/]+", 35: ".+"}), topic))
        return ret


# Manually create Mock-Object


class Options():

    def __init__(self):
        self.mqtt_broker_host = "192.168.88.229"
        self.mqtt_broker_port = 1883
        self.mqtt_client_id = "testing"


async def main():
    app = {}
    options = Options()
    app['options'] = options

    mqttc = Mqtt(app)
    await mqttc.start_async()

    mqttc.client.publish('cmnd/dev01/POWER1', payload="ON", qos=1)

    try:
        await asyncio.sleep(60)
    except asyncio.CancelledError:
        print('MQTT: trapped cancel')
    finally:
        await mqttc.client.disconnect()


if __name__ == "__main__":
    asyncio.run(main())



import asyncssh
import asyncio
import aiohttp
from aiohttp import ClientError


from logging import getLogger
_log = getLogger(__name__)

_instance = None


def get_cloud():
    global _instance
    return _instance


def setup_cloud(app=None, options=None, parameters=None):
    global _instance
    if _instance is not None:
        del _instance
    _instance = Cloud(app=app, options=options, parameters=parameters)
    return _instance


class Cloud:
    def __init__(self, app=None, options=None, parameters=None):
        global _instance
        if _instance is None:
            _instance = self

        self._parameters = parameters
        self._options = options
        self._STOP = False
        self._conn = None
        self._listener = None
        app.cleanup_ctx.append(self._ssh_engine)

    async def _ssh_engine(self, app):
        task = asyncio.create_task(self._startup(app))
        yield
        await self._shutdown(app, task)

    async def _startup(self, app) -> None:
        _log.debug('Ssh Engine startup called')

# TODO move the following code to Web-Interface

        while not self._STOP:
            if (self._parameters.ssh_server_hostname is None or
                self._parameters.ssh_server_port is None or
                self._parameters.ssh_server_listen_port is None or
                self._parameters.cloud_server_url is None or
                    self._parameters.ssh_user_key is None):
                await asyncio.sleep(3)
                continue
            ssh_user_key = asyncssh.import_private_key(
                self._parameters.ssh_user_key)
            try:
                async with asyncssh.connect(
                        self._parameters.ssh_server_hostname,
                        port=self._parameters.ssh_server_port,
                        username=self._parameters.cloud_identifier,
                        client_keys=ssh_user_key,
                        known_hosts=self._options.ssh_known_hosts,
                        keepalive_interval=60
                ) as self._conn:
                    self._listener = await self._conn.forward_remote_port(
                        'localhost',  # Bind to localhost on remote Server
                        self._parameters.ssh_server_listen_port,
                        'localhost',
                        self._options.http_server_bind_port
                    )
                    _log.debug('remote 127.0.0.1:{}, local 127.0.0.1:{}'.format(  # noqa: E501
                        self._parameters.ssh_server_listen_port,
                        self._options.http_server_bind_port))
                    await self._listener.wait_closed()
            except Exception as e:
                _log.debug(f'SSH-Exception: {e}')
            await asyncio.sleep(10)

    async def _shutdown(self, app, task) -> None:
        _log.debug('Ssh Engine shutdown called')
        self._STOP = True
        if self._listener is not None:
            self._listener.close()
        if self._conn is not None:
            self._conn.close()
#            await self._conn.wait_closed()
        return None

    async def create_cloud(self):
        key = asyncssh.generate_private_key(self._options.ssh_key_algorithm)
        self._parameters.ssh_user_key = key.export_private_key()
        public_user_key = key.export_public_key().decode('ascii')

        data = {'cloud_identifier': self._parameters.cloud_identifier,
                'public_user_key': public_user_key}

        try:
            async with aiohttp.ClientSession() as session:
                async with session.post(self._options.cloud_api_url,
                                        json=data) as resp:
                    if resp.status != 200:
                        return
                    data = await resp.json()
        except ClientError as e:
            _log.debug(f'Unable to access remote api: {e}')
            return False
        else:
            self._parameters.ssh_server_hostname = data['ssh_server_hostname']  # noqa: E501
            self._parameters.ssh_server_port = data['ssh_server_port']
            self._parameters.ssh_server_listen_port = data['ssh_server_listen_port']  # noqa: E501
            self._parameters.cloud_server_url = data['cloud_server_url']  # noqa: E501
            return True

    async def check_cloud(self):
        if (
            self._parameters.ssh_server_hostname is None or
            self._parameters.ssh_server_port is None or
            self._parameters.ssh_server_listen_port is None or
            self._parameters.ssh_user_key is None or
            self._parameters.cloud_server_url is None
        ):
            return None
        else:
            url = '{}/{}/'.format(
                       self._parameters.cloud_server_url,
                       self._parameters.cloud_identifier)
            return url



from sqlite3 import IntegrityError, Row
from riego.db import get_db

import aiohttp_jinja2
from aiohttp import web
from aiohttp_session import get_session, new_session

import bcrypt
import secrets
import json
import asyncio

from logging import getLogger
_log = getLogger(__name__)


router = web.RouteTableDef()


def setup_routes_security(app):
    app.add_routes(router)


@router.get("/login", name='login')
@aiohttp_jinja2.template("security/login.html")
async def login(request: web.Request):
    redirect = request.rel_url.query.get("redirect", "")
    csrf_token = secrets.token_urlsafe()
    session = await new_session(request)
    # session = await get_session(request)
    session['csrf_token'] = csrf_token
    session.changed()
    return {'csrf_token': csrf_token, 'redirect': redirect}


@ router.post("/login")
async def login_apply(request: web.Request):
    form = await request.post()
    session = await get_session(request)
    if session.get('csrf_token') != form['csrf_token']:
        # Normally not possible
        await asyncio.sleep(2)
        return web.HTTPUnauthorized()

    if form.get('identity') is None:
        await asyncio.sleep(2)
        return web.HTTPSeeOther(request.app.router['login'].url_for())

    cursor = get_db().conn.cursor()
    cursor.execute("""SELECT *, 'login' AS provider
                    FROM users
                    WHERE identity = ?""", (form['identity'],))
    user = cursor.fetchone()

    if (
        user is None or
        user['is_disabled'] or
        user['password'] is None or
        not len(user['password'])
    ):
        await asyncio.sleep(2)
        return web.HTTPSeeOther(request.app.router['login'].url_for())
    if not bcrypt.checkpw(form['password'].encode('utf-8'),
                          user['password'].encode('utf-8')):
        await asyncio.sleep(2)
        return web.HTTPSeeOther(request.app.router['login'].url_for())

    session['user_id'] = user['id']

    location = form.get('redirect')
    if location is None or location == '':
        location = request.app.router['home'].url_for()
    response = web.HTTPSeeOther(location=location)
# TODO use create_remember_me
    if form.get('remember_me') is not None:
        remember_me = secrets.token_urlsafe()
        try:
            with get_db().conn:
                get_db().conn.execute(
                    '''UPDATE users
                       SET remember_me = ?
                       WHERE id = ? ''',
                    (remember_me, user['id']))
        except IntegrityError:
            pass
        response.set_cookie("remember_me", remember_me,
                            max_age=request.app['options'].max_age_remember_me,
                            httponly=True,
                            samesite='strict')
    return response


@ router.get("/logout", name='logout')
async def logout(request: web.Request):
    user = await get_user(request)
    if user is not None:
        await delete_websocket_auth(request, user=user)
# TODO use delete_remember_me
        try:
            with get_db().conn:
                get_db().conn.execute("""UPDATE users
                                        SET remember_me = ''
                                        WHERE id = ?""", (user['id'],))
        except IntegrityError:
            pass
    session = await get_session(request)
    if session is not None:
        session.pop('user_id', None)
    response = web.HTTPSeeOther(request.app.router['login'].url_for())
#    response.set_cookie('remember_me', None,
#                        expires='Thu, 01 Jan 1970 00:00:00 GMT')
    response.del_cookie('remember_me')
    return response


@ router.get("/profile", name='profile')
@ aiohttp_jinja2.template("security/profile.html")
async def profile(request: web.Request):
    return {}


@ router.post("/profile")
async def profile_apply(request: web.Request):
    form = await request.post()
    user = await get_user(request)

# TODO check old_password and equality of pw1 an pw2
    password = form['new_password_1'].encode('utf-8')
    password = bcrypt.hashpw(password, bcrypt.gensalt())
    password = password.decode('utf-8')
    try:
        with get_db().conn:
            get_db().conn.execute(
                '''UPDATE users
                    SET password = ?
                    WHERE id = ? ''', (password, user['id']))
    except IntegrityError:
        return web.HTTPSeeOther(request.app.router['passwd'].url_for())

    return web.HTTPSeeOther(request.app.router['home'].url_for())
    return {}  # not reached


async def current_user_ctx_processor(request):
    websocket_auth = '''{"token": "", "sequence": ""}'''
    user = await get_user(request)
    if user is not None:
        websocket_auth = await create_websocket_auth(request, user=user)
    return {'user': user, 'websocket_auth': json.loads(websocket_auth)}


async def get_user(request) -> Row:
    session = await get_session(request)
    db = get_db()
    user_id = session.get('user_id')
    if user_id is not None:
        cursor = db.conn.cursor()
        cursor.execute("""SELECT *, 'login' AS provider
                          FROM users
                          WHERE id = ?""", (user_id,))
        user = cursor.fetchone()
        if user is None or user['is_disabled']:
            session.pop('user_id', None)
            return None
        return user

    return await check_remember_me_auth(request)


async def check_permission(request, permission=None) -> Row:
    user = await get_user(request)
    db = get_db()
    if user is None:
        return None
    if user['is_disabled']:
        return None
    if user['is_superuser']:
        return user
    if permission is None:
        return user

    cursor = db.conn.cursor()
    cursor.execute(
        'SELECT * FROM users_permissions WHERE user_id = ?', (user['id'],))
    for row in cursor:
        if row['name'] == permission:
            return user

    return None


async def raise_permission(request: web.BaseRequest, permission: str = None):
    """Generate redirection to login form if permission is not
    sufficent. Append query string with information for redirecting
    after login to the original url.

    :param request: [description]
    :type request: web.Baserequest
    :param permission: If no permission is given, check auth only
    :type permission: str, optional
    :raises web.HTTPSeeOther: [description]
    """
    if await check_permission(request, permission=permission) is None:
        raise web.HTTPSeeOther(
            request.app.router['login'].url_for(
            ).with_query(
                {"redirect": str(request.rel_url)}
            )
        )


async def create_websocket_auth(request: web.BaseRequest,
                                user: Row = None) -> json:
    """create token and sequence number if not exist in session. Than
    a) put token and sequence into session for using in templates
    b) put hashed token and sequence into database for later checking against
       data recived with websocket.py

    :param request: [description]
    :type request: web.BaseRequest
    :param user: [description], defaults to None
    :type user: Row, optional
    :return: [description]
    :rtype: json
    """
    if user is None:
        return None
    db = get_db()
    session = await get_session(request)
    session_key = request.app['options'].session_key_websocket_auth
    websocket_auth = session.get(session_key, '')
    if len(websocket_auth) > 0:
        return websocket_auth

    sequence = secrets.token_urlsafe()
    token = secrets.token_urlsafe()
    token_hash = bcrypt.hashpw(token.encode('utf-8'), bcrypt.gensalt())

    try:
        with db.conn:
            db.conn.execute('''INSERT INTO users_tokens
                                (sequence, hash, category, user_id)
                                VALUES (?, ?, ?, ?)''',
                            (sequence, token_hash, "websocket", user['id']))
    except IntegrityError as e:
        _log.error(f'Unable to insert token: {e}')
        return None
    websocket_auth = json.dumps({"sequence": sequence, "token": token})
    session[session_key] = websocket_auth
    return websocket_auth


async def delete_websocket_auth(request, user=None):
    db = get_db()
    session = await get_session(request)
    session_key = request.app['options'].session_key_websocket_auth
    if session is not None:
        session.pop(session_key, None)
    try:
        with db.conn:
            db.conn.execute('''DELETE FROM users_tokens
                               WHERE category = ? AND user_id = ?''',
                            ("websocket", user['id']))
    except IntegrityError:
        pass
    return None


async def create_remember_me_auth(request, response=None, user: Row = None):
    remember_me = secrets.token_urlsafe()
    try:
        with get_db().conn:
            get_db().conn.execute(
                '''UPDATE users
                    SET remember_me = ?
                    WHERE id = ? ''',
                (remember_me, user['id']))
    except IntegrityError:
        return None
    response.set_cookie("remember_me", remember_me,
                        max_age=request.app['options'].max_age_remember_me,
                        httponly=True,
                        samesite='strict')
    return response


async def delete_remember_me_auth(request, response=None, user: Row = None):
    if user is not None:
        try:
            with get_db().conn:
                get_db().conn.execute("""UPDATE users
                                        SET remember_me = ''
                                        WHERE id = ?""", (user['id'],))
        except IntegrityError:
            pass

    session = await get_session(request)
    session.pop('user_id', None)
#    response.set_cookie('remember_me', None,
#                        expires='Thu, 01 Jan 1970 00:00:00 GMT')
    response.del_cookie('remember_me')
    return response


async def check_remember_me_auth(request) -> Row:
    db = get_db()
    user = None
    remember_me = request.cookies.get('remember_me')
    if remember_me is not None:
        cursor = db.conn.cursor()
        cursor.execute("""SELECT *, 'cookie' AS provider
                           FROM users
                           WHERE remember_me = ?""", (remember_me,))
        user = cursor.fetchone()
        if user is not None and user['is_disabled']:
            try:
                with db.conn:
                    db.conn.execute("""UPDATE users
                                       SET remember_me = ''
                                       WHERE id = ?""", (user['id'],))
            except IntegrityError:
                pass
    return user



import setuptools
from riego.__init__ import __version__


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="riego",
    version=__version__,
    author="Stephan Winter",
    author_email="riego@finca-panorama.es",
    url="https://github.com/py-steph/riego",
    description="Watering System",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=[
        'gmqtt',
        'ConfigArgParse',
        'yoyo-migrations',
        'aiohttp',
        'aiohttp_jinja2',
        'jinja2',
        'uvloop; sys_platform != "win32"',
        'cchardet',
        'aiohttp_debugtoolbar',
        'aiohttp_session[secure]',
        'bcrypt',
        'asyncssh',
    ],
    include_package_data=True,
    scripts=['bin/setup_riego_service.sh'],
    entry_points={
        'console_scripts': ['riego=riego.app:main'],
    },
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 3 - Alpha",
        "Framework :: AsyncIO",
        "Operating System :: POSIX :: Linux",
    ],
    python_requires='>=3.7',
    zip_safe=False,
)
